/************************************************************************
 *   MODULE  : VIP_MASK.C
 *
 *   PURPOSE : Fills the visibility probability product with the
 *             appropriate values.
 *
 *   AUTHOR  : Ron Hawkins & Craig Woodfolk, SAIC (original)
 *             Kevin Russo, SAIC (1, 3)
 *             Melissa Mathews, SAIC (2)
 *
 *   Change     Date          Reason
 *   ------     ----          ------
 *   Original   25 Sep 92     n/a
 *   1          20 Oct 92     overall program update
 *   2          17 Jun 93     converted XVT 2.12 calls to XVT 3.01
 *   3          21 Jul 93     implemented new TEC LOS algorithm
 *
 ************************************************************************/

/****************
 * header files *
 ****************/

#include "los.h"

/************************************************************************
 *   FUNCTION     : vip_mask
 *
 *   PURPOSE      : visibility probability mask building routine
 *
 *   INPUTS       : elev_hgts      - elevation buffer
 *                  prod_buff_ydim - product buffer (lat)
 *                  prod_buff_xdim - product buffer (lon)
 *                  profile_color  - profile color buffer
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : make_one_radial in module MAKE_RDL.C
 *
 *   CALLS        : none
 *
 ************************************************************************/

#ifdef PROTO
void vip_mask( short pHMEM elev_hgts, long pHMEM prod_buff_ydim,
               long pHMEM prod_buff_xdim, char *pHMEM profile_color )
#else
void vip_mask( elev_hgts, prod_buff_ydim, prod_buff_xdim, profile_color )
short  pHMEM elev_hgts;
long   pHMEM prod_buff_ydim;
long   pHMEM prod_buff_xdim;
char   *pHMEM profile_color;
#endif
{

   long    i, index;          /* counter                                */
   GFLOAT  distance;          /* distance to current point along radial */
   GFLOAT  vert_diff;         /* antenna top and target top difference  */
   GFLOAT  vert_accuracy;     /* absolute vertical accuracy from DTED   */
   GFLOAT  max_los_ratio;     /* maximum LOS ratio                      */
   GFLOAT  tgt_ratio;         /* antenna to target ratio                */
   GFLOAT  los_ratio;         /* antenna to current LOS ratio           */
   GFLOAT  new_ratio;         /* antenna to adjusted LOS ratio          */
   GFLOAT  correction;        /* elevation correction                   */
   GFLOAT  eff_radius;        /* effective Earth radius                 */
   GFLOAT  acuity_distance;   /* target acuity distance                 */

   char pHMEM prfl_col;       /* profile color pointer                  */

   BYTE  vis, hid, prvis, prhid, maxcheck;

   extern  PARAM_CALC      param;           /* calculations location    */
   extern  SETTINGS_TYPE   settings;        /* user input parameters    */
   extern  RASTER          *prod_raster;    /* product raster           */
   extern  char pHMEM      prod1_ptr;       /* product data pointer     */
   extern  PREFS_TYPE      sys_prefs;       /* System preferences       */

   extern BYTE pHMEM      num_hidden;
   extern BYTE pHMEM      num_visible;
   extern BYTE pHMEM      num_prhidden;
   extern BYTE pHMEM      num_prvisible;

   /* set local profile color pointer for more clarity */
   if (settings.profile)
      prfl_col = (char pHMEM)*profile_color;

   max_los_ratio   = INIT_LOS_VAL;
   vert_accuracy   = settings.vert_acc;
   acuity_distance = settings.target.width / ACUITY_FACTOR;
   eff_radius      = (GFLOAT)2 * EARTH_RADIUS;

   for (i = ONE; i < param.nhts_in_radial; i++)
   {
      if (elev_hgts[i] < (short)ZERO)
         continue;

      index = prod_buff_ydim[i] * prod_raster->data_matrix.width +
              prod_buff_xdim[i];

      distance   = i * param.radial_step_size;
      correction = (GFLOAT) elev_hgts[i] - ((distance * distance) / eff_radius);
      vert_diff  = correction - (GFLOAT)elev_hgts[0] -
                   (GFLOAT) settings.observer.height +
                   (GFLOAT) settings.target.height;
      tgt_ratio  = vert_diff / distance;
      los_ratio  = (vert_diff - (GFLOAT)settings.target.height) / distance;

      if (los_ratio > max_los_ratio)
         max_los_ratio = los_ratio;

      if ((tgt_ratio < max_los_ratio) ||
          ((sys_prefs.acuity == TRUE) && (distance > acuity_distance)))
      {
         new_ratio = (vert_diff + vert_accuracy) / distance;

         if ((sys_prefs.acuity == TRUE) && (distance > acuity_distance))
         {
            prod1_ptr[index] = (char)BEYOND_ACUITY;
         }
         else if (new_ratio < max_los_ratio)
         {
            /* priority mask is VISIBILITY */
            num_hidden[index]++;
         }
         else
         {
            num_prhidden[index]++;
         }
      }
      else
      {
         /* tgt_ratio > max_los_ratio */
         new_ratio = (vert_diff - vert_accuracy) / distance;

         if (new_ratio < max_los_ratio)
         {
            num_prvisible[index]++;
         }
         else
         {
            /* priority mask is VISIBILITY */
            num_visible[index]++;
         }
      }
   } /* end of param.nhts_in_radial loop */


   /* assign color codes to profile array */
   if (settings.profile)
   {
      for (i = ONE; i < param.nhts_in_radial; i++)
      {
         index = prod_buff_ydim[i] * prod_raster->data_matrix.width +
                 prod_buff_xdim[i];

         if (elev_hgts[i] < (short)ZERO)
         {
            prfl_col[i] = (char)ZERO;
         }
         else if (prod1_ptr[index] == (char)BEYOND_ACUITY)
         {
            prfl_col[i] = (char)BEYOND_ACUITY;
         }
         else
         {
            hid = num_hidden[index];

            vis = num_visible[index];

            prhid = num_prhidden[index];

            prvis = num_prvisible[index];

            if ((hid == (BYTE)ZERO) && (vis == (BYTE)ZERO) &&
                (prhid == (BYTE)ZERO) && (prvis == (BYTE)ZERO))
               continue;

            if (prod1_ptr[index] != (char)ZERO)
               continue;

            maxcheck = vis;
            maxcheck = (BYTE) maxof (maxcheck, hid);
            maxcheck = (BYTE) maxof (maxcheck, prvis);
            maxcheck = (BYTE) maxof (maxcheck, prhid);

            if (maxcheck == vis)
            {
               prod1_ptr[index] = (char)DEF_VISIBLE;

               prfl_col[i] = (char)DEF_VISIBLE;
            }
            else if (maxcheck == hid)
            {
               prod1_ptr[index] = (char)DEF_HIDDEN;

               prfl_col[i] = (char)DEF_HIDDEN;
            }
            else if (maxcheck == prvis)
            {
               prod1_ptr[index] = (char)PROB_VISIBLE;

               prfl_col[i] = (char)PROB_VISIBLE;
            }
            else
            {
               prod1_ptr[index] = (char)PROB_HIDDEN;

               prfl_col[i] = (char)PROB_HIDDEN;
            }
         }
      }
   }
}
