/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


#include <Map/mapfeature.h>
#include <Map/mapprim.h>
#include <Map/mapproject.h>

#include <TopoFace/topoedge.h>
#include <TopoFace/topoface.h>
#include <TopoFace/toponode.h>

#include <Unidraw/iterator.h>
#include <string.h>

/*****************************************************************************/

MapPrimitive::MapPrimitive(MapFeature* feature) {
    _feature = feature;
    _topo = nil;
}

MapPrimitive::~MapPrimitive() {
    delete _topo;
} 

void MapPrimitive::feature(MapFeature* feature) {
    _feature = feature;
}

MapFeature* MapPrimitive::feature() const {
    return _feature;
}

int MapPrimitive::npts() const { return topo()->npts(); }
const float* MapPrimitive::xpoints() const { return topo()->xpoints(); }
const float* MapPrimitive::ypoints() const { return topo()->ypoints(); }
const float* MapPrimitive::zpoints() const { return topo()->zpoints(); }

void MapPrimitive::origin(float& x, float& y, float& z) const {
    x = topo()->xpoints()[0];
    y = topo()->ypoints()[0];
    z = topo()->zpoints() ? topo()->zpoints()[0] : 0.0;
}

float MapPrimitive::xorigin() const { return topo()->xpoints()[0]; }
float MapPrimitive::yorigin() const { return topo()->ypoints()[0]; }
float MapPrimitive::zorigin() const { return topo()->zpoints() ? topo()->zpoints()[0] : 0.0; }

float MapPrimitive::xextent() const { return topo()->xmax() - topo()->xmin(); }
float MapPrimitive::yextent() const { return topo()->ymax() - topo()->ymin(); }
float MapPrimitive::zextent() const { return topo()->zmax() - topo()->zmin(); }

void MapPrimitive::insert_points(int npts, float* x, float* y, float* z) {
    topo()->insert_points(npts, x, y, z);
    return;
}

void MapPrimitive::insert_points(int npts, int* x, int* y, int* z) {
    topo()->insert_points(npts, x, y, z);
    return;
}

void MapPrimitive::insert_pointers(int npts, float* x, float* y, float*z, boolean freeflag) {
    topo()->insert_pointers(npts, x, y, z, freeflag);
    return;
}

/*****************************************************************************/

MapEdge::MapEdge(MapFeature* feature, boolean topoflag) : MapPrimitive(feature) { 
    if (topoflag) _topo = new TopoEdge(this); 
}

MapFace* MapEdge::left_face() const { return (MapFace*)topo()->left_face()->value(); }
MapFace* MapEdge::right_face() const { return (MapFace*)topo()->right_face()->value(); }

MapNode* MapEdge::start_node() const { return (MapNode*)topo()->start_node()->value(); }
MapNode* MapEdge::end_node() const { return (MapNode*)topo()->end_node()->value(); }

MapEdge* MapEdge::next_edge(MapFace* face, MapNode* node) const {
    return nil;
}

boolean MapEdge::forward(MapNode* node) const { 
    return nil;
}

void MapEdge::attach_nodes(MapNode* start, MapNode* end) {
    topo()->attach_nodes(start->topo(), end->topo());
    return;
}

void MapEdge::attach_faces(MapFace* left, MapFace* right) {
    topo()->attach_faces(left ? left->topo() : nil, right ? right->topo() : nil);
    return;
}

/*****************************************************************************/

MapEdgeList::MapEdgeList(MapFeature* feature) : MapPrimitive(feature) {}

void MapEdgeList::first(Iterator& i) const { topo()->first(i); }
void MapEdgeList::last(Iterator& i) const { topo()->last(i); }
void MapEdgeList::next(Iterator& i) const { topo()->next(i); }
void MapEdgeList::prev(Iterator& i) const { topo()->prev(i); }
boolean MapEdgeList::done(Iterator i) const { return topo()->done(i); }
boolean MapEdgeList::is_empty() const { return topo()->is_empty(); }
int MapEdgeList::number() const { return topo()->number(); }

void MapEdgeList::append(MapEdge* edge) {
    topo()->append(edge->topo());
}

void MapEdgeList::prepend(MapEdge* edge) {
    topo()->prepend(edge->topo());
}

/*****************************************************************************/

MapFace::MapFace(MapFeature* feature, boolean topoflag) : MapEdgeList(feature) {
    if (topoflag) _topo = new TopoFace(this); 
}

int MapFace::npts() const { return topo()->npts(); }

boolean MapFace::clockwise(MapEdge* node) const { 
    return topo()->clockwise(node->topo());
}

/*****************************************************************************/

MapNode::MapNode(MapFeature* feature, boolean topoflag) : MapEdgeList(feature) {
    if (topoflag) _topo = new TopoNode(this); 
}

/*****************************************************************************/

MapPoint::MapPoint(MapFeature* feature, boolean topoflag) : MapEdge(feature, topoflag) {}

MapFace* MapPoint::containing_face() const { 
    return (MapFace*)topo()->left_face()->value(); 
}

/*****************************************************************************/

MapText::MapText(MapFeature* feature, boolean topoflag)
: MapPoint(feature, topoflag)
{
    _text = nil;
}

MapText::~MapText() {
    if (_text)
	delete [] _text;
}

const char* MapText::text() const { return _text; }

void MapText::text(const char* txt) {
    if (_text)
	delete [] _text;
    _text = strdup(txt);
}

/*****************************************************************************/

MapRaster::MapRaster(MapFeature* feature, boolean topoflag) : MapFace(feature, topoflag) {}

/*****************************************************************************/

MapDem::MapDem(MapFeature* feature, boolean topoflag) : MapFace(feature, topoflag) {}

