/*
 * Copyright 1993-1995 Cartoactive Systems, Vectaport Inc., Cider Press
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <VpfUnidraw/vpfclasses.h>
#include <VpfUnidraw/vpfcmds.h>
#include <VpfUnidraw/vpfcomps.h>
#include <VpfUnidraw/vpfeditor.h>
#include <VpfUnidraw/vpfedtile.h>
#include <VpfUnidraw/vpfviewstate.h>

#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapcmds.h>
#include <MapUnidraw/mapdialogs.h>
#include <MapUnidraw/mapproj.h>
#include <MapUnidraw/mapstates.h>

#include <Vpf/vpfdb.h>
#include <Vpf/vpfcov.h>
#include <Vpf/vpffclass.h>
#include <Vpf/vpffeature.h>
#include <Vpf/vpflib.h>
#include <Vpf/vpftiledfc.h>

#include <Map/mapprim.h>

#include <OverlayUnidraw/ovcomps.h>
#include <OverlayUnidraw/ovpage.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <Unidraw/Commands/edit.h>

#include <Unidraw/Graphic/ellipses.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Graphic/polygons.h>
#include <Unidraw/Graphic/pspaint.h>

#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/globals.h>
#include <Unidraw/grid.h>
#include <Unidraw/iterator.h>
#include <Unidraw/selection.h>
#include <Unidraw/viewer.h>
#include <Unidraw/ulist.h>
#include <Unidraw/unidraw.h>

#include <InterViews/cursor.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>

#include <limits.h>
#include <string.h>

/*****************************************************************************/

static const float GRID_XINCR = 8;                 // default grid spacing
static const float GRID_YINCR = 8;

ClassId VpfOpenCmd::GetClassId () { return VPFOPEN_CMD; }

boolean VpfOpenCmd::IsA (ClassId id) {
    return VPFOPEN_CMD == id || MapOpenCmd::IsA(id);
}

VpfOpenCmd::VpfOpenCmd (ControlInfo* c, OpenFileChooser* fc) : MapOpenCmd(c, fc) { }
VpfOpenCmd::VpfOpenCmd (Editor* ed, OpenFileChooser* fc) : MapOpenCmd(ed, fc) { }

Command* VpfOpenCmd::Copy () {
    Command* copy = new VpfOpenCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void VpfOpenCmd::Execute () {
    MapEditor* ed = (MapEditor*)GetEditor();
    VpfIdrawComp* orig_comp = (VpfIdrawComp*)ed->GetComponent();

    //open up
    OvViewCompCmd::Execute();
    VpfIdrawComp* comp = (VpfIdrawComp*)ed->GetComponent();
    if (comp != orig_comp) {
        Grid* grid = ed->GetViewer()->GetGrid();
        grid->SetSpacing(GRID_XINCR, GRID_YINCR);
    }

    MapProjectionCmd* pcmd = new MapProjectionCmd(ed);
    pcmd->SetProjection(comp->GetProjectionName());

    //single database case, initialize
    Iterator i;
    comp->First(i);    
    VpfDbComp* dbcomp = (VpfDbComp*)comp->GetComp(i);
    VpfDatabase* db = dbcomp->GetDatabase();

    VpfDatabase* orig_db = (VpfDatabase*)ed->GetMapDatabase();
    if (orig_db && strcmp(orig_db->path(), db->path()) == 0) {
	dbcomp->SetDatabase(orig_db);
    } else {
	db->init();
    }   
    dbcomp->init();
    ed->SetMapDbComp(dbcomp);
    ed->OpenInit();

    MapShowAllCmd* showall = new MapShowAllCmd(ed);

    Attribute* imattr;
    if (imattr = comp->GetAttributeList()->GetAttr("initmag")) {
      showall->SetMag(imattr->Value()->float_val());
    }
    else {
      Catalog* catalog = unidraw->GetCatalog();
      char* mag = strdup(catalog->GetAttribute("initmag"));
      double fmag = atof(mag);
      showall->SetMag(fmag);
    }

    showall->Execute();
}

/*****************************************************************************/

ClassId VpfShowTilesCmd::GetClassId () { return VPFSHOWTILES_CMD; }

boolean VpfShowTilesCmd::IsA (ClassId id) {
    return VPFSHOWTILES_CMD == id || Command::IsA(id);
}

VpfShowTilesCmd::VpfShowTilesCmd (ControlInfo* c) : Command(c) { }
VpfShowTilesCmd::VpfShowTilesCmd (Editor* ed) : Command(ed) { }

Command* VpfShowTilesCmd::Copy () {
    Command* copy = new VpfShowTilesCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void VpfShowTilesCmd::Execute () {
    MapEditor* editor = (MapEditor*)GetEditor();
    MapViewState* view = editor->GetMapViewState();
    MapDatabase* mapdb = editor->GetMapDatabase();
    MapLibrary* maplib;
    if (!editor->libstate()->maplib() || strlen(editor->libstate()->maplib())==0)
	maplib = choosetiledlib(mapdb);
    else {
	maplib = (MapLibrary*)mapdb->library(editor->libstate()->maplib());
	if (!maplib || !maplib->tiled())
	    maplib = choosetiledlib(mapdb);
    }
    int i;
    if (maplib == nil) return;
	editor->libstate()->maplib(maplib->name());
    MapCoverage* mapcov = (MapCoverage*)maplib->coverage("tileref");
    if (mapcov == nil) return;
	editor->covstate()->mapcov(mapcov->name(), mapcov->description());
    MapFeatureClass* mapfc = (MapFeatureClass*)mapcov->feature_class("tileref");
    if (mapfc == nil) return;
	editor->fclassstate()->mapfclass(mapfc->name(), mapfc->description());

    MapShowCmd* showcmd = new MapShowCmd(editor);
    showcmd->Execute();
}

boolean VpfShowTilesCmd::Reversible () { return true; }

/*****************************************************************************/

ClassId VpfToolbarCmd::GetClassId () { return VPFTOOLBAR_CMD; }

boolean VpfToolbarCmd::IsA (ClassId id) {
    return VPFTOOLBAR_CMD == id || Command::IsA(id);
}

VpfToolbarCmd::VpfToolbarCmd (ControlInfo* c, boolean tiled)
: Command(c)
{
    _tiled = tiled;
}
VpfToolbarCmd::VpfToolbarCmd (Editor* ed, boolean tiled)
: Command(ed)
{
    _tiled = tiled;
}

Command* VpfToolbarCmd::Copy () {
    Command* copy = new VpfToolbarCmd(CopyControlInfo(), _tiled);
    InitCopy(copy);
    return copy;
}

void VpfToolbarCmd::Execute () {
    VpfEditor* editor = (VpfEditor*)GetEditor();
    if (_tiled)
	editor->toolbar0();
    else
	editor->toolbar1();
}

boolean VpfToolbarCmd::Reversible () { return false; }

/*****************************************************************************/

ClassId VpfUpdateMenusCmd::GetClassId () { return VPFUPDATEMENUS_CMD; }

boolean VpfUpdateMenusCmd::IsA (ClassId id) {
    return VPFUPDATEMENUS_CMD == id || Command::IsA(id);
}

VpfUpdateMenusCmd::VpfUpdateMenusCmd (ControlInfo* c)
: Command(c)
{
}

VpfUpdateMenusCmd::VpfUpdateMenusCmd (Editor* ed)
: Command(ed)
{
}

Command* VpfUpdateMenusCmd::Copy () {
    Command* copy = new VpfUpdateMenusCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void VpfUpdateMenusCmd::Execute () {
    VpfEditor* editor = (VpfEditor*)GetEditor();
    if (editor) {
      editor->UpdateLibraryMenu();
      editor->UpdateLayerMenu();
    }
}

boolean VpfUpdateMenusCmd::Reversible () { return false; }

/*****************************************************************************/

VpfGraticuleCmd::VpfGraticuleCmd (ControlInfo* c)
: Command(c)
{
}

VpfGraticuleCmd::VpfGraticuleCmd (Editor* ed)
: Command(ed)
{
}

void VpfGraticuleCmd::Execute() {
  VpfEditor* ed = (VpfEditor*)GetEditor();
  MapDbComp* mapdbcomp = ed->GetMapDbComp();
  if (mapdbcomp && !ed->Graticule()) {
    ed->GetWindow()->cursor(hourglass);
    int interval = 15;
    int line_interval = 5;

    MapFeature** feats = new MapFeature*[(360/interval)+1 + (180/interval)+1];
    int fi = 0;
    for (int i = -180; i <= 180; i += interval) {
      int npts = (180 / line_interval) + 1;
      float* x = new float[npts];
      float* y = new float[npts];
      int ip = 0;
      for (int j = -90; j <= 90; j += line_interval) {
	x[ip] = i;
	y[ip++] = j;
      }
      MapFeature* f = new MapFeature(-1, MapFeature::LineType);
      MapEdge* line = new MapEdge(f);
      line->insert_pointers(npts, x, y);
      f->edge_primitive(line);
      feats[fi++] = f;
    }
    for (int i = -90; i <= 90; i += interval) {
      int npts = (360 / line_interval) + 1;
      float* x = new float[npts];
      float* y = new float[npts];
      int ip = 0;
      for (int j = -180; j <= 180; j += line_interval) {
	x[ip] = j;
	y[ip++] = i;
      }
      MapFeature* f = new MapFeature(-1, MapFeature::LineType);
      MapEdge* line = new MapEdge(f);
      line->insert_pointers(npts, x, y);
      f->edge_primitive(line);
      feats[fi++] = f;
    }
    MapFeatureClass* fc = new MapFeatureClass(fi, feats, "graticule");
    MapFClassComp* fccomp = new MapFClassComp(fc);
    for (int i = 0; i < fi; i++) {
      feats[i]->feature_class(fc);
      MapFeatureComp* fcomp = new MapFeatureComp(nil, fccomp);
      fcomp->feature(feats[i]);
      fccomp->Append(fcomp);
    }
    ed->Graticule(fccomp);
    Iterator it;
    mapdbcomp->First(it);
    MapLibComp* maplibcomp = (MapLibComp*)mapdbcomp->GetComp(it);
    maplibcomp->First(it);
    MapCovComp* mapcovcomp = (MapCovComp*)maplibcomp->GetComp(it);
    mapcovcomp->Append(fccomp);
    mapcovcomp->Notify();
    unidraw->Update();
  }
}

Command* VpfGraticuleCmd::Copy () {
    Command* copy = new VpfGraticuleCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}
