/*
 * Copyright 1994, 1995 Vectaport Inc., Cider Press
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Map/mapprim.h>

#include <Vpf/vpfcov.h>
#include <Vpf/vpfdb.h>
#include <Vpf/vpflib.h>
#include <Vpf/vpffclass.h>
#include <Vpf/vpftiledfc.h>
#include <Vpf/vpffeature.h>

#include <VpfUnidraw/vpfclasses.h>
#include <VpfUnidraw/vpfcomps.h>
#include <VpfUnidraw/vpfeditor.h>
#include <VpfUnidraw/vpfscripts.h>
#include <VpfUnidraw/vpfviewstate.h>

#include <MapUnidraw/mapcmds.h>
#include <MapUnidraw/mapcomps.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <OverlayUnidraw/ovdamage.h>
#include <OverlayUnidraw/ovpage.h>
#include <OverlayUnidraw/paramlist.h>

#include <Unidraw/Commands/command.h>
#include <Unidraw/catalog.h>
#include <Unidraw/iterator.h>
#include <Unidraw/selection.h>
#include <Unidraw/ulist.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>

#include <InterViews/cursor.h>
#include <InterViews/window.h>

#include <stdlib.h>
#include <string.h>

#define EPS10 1.0e-10

void projectpts(PJ* proj, int npts, const float* fx, const float* fy, float* ix, float* iy, float scale = 1.0) {
    UV data;
    for (int i=0; i<npts; i++) {
	if (proj) {
	    data.u = fx[i] * DEG_TO_RAD;
	    data.v = fy[i] * DEG_TO_RAD;
	    data = pj_fwd(data, proj);
	    if (data.u == HUGE_VAL) {
		/* cerr << "projection error\n"; */
		data.u = fx[i] * DEG_TO_RAD;
		data.v = fy[i] * DEG_TO_RAD;
		if (fx[i]==90.0)
		    data.u -= 2 * EPS10;
		if (fx[i]==-90.0)
		    data.u += 2 * EPS10;
		if (fy[i]==90.0)
		    data.v -= 2 * EPS10;
		if (fy[i]==-90.0)
		    data.v += 2 * EPS10;
		data = pj_fwd(data, proj);
		//		if (data.u == HUGE_VAL)
		//		    cerr << "projection error again\n";
	    }
	    ix[i] = data.u;
	    iy[i] = data.v;
	} else {
	    ix[i] = fx[i] * scale;
	    iy[i] = fy[i] * scale;
	}
    }
}

void project_feature(PJ* pj, MapFeature* mf) {
  switch(mf->type()) {
  case MapFeature::PointType:
    {
      MapPoint* point = mf->point_primitive();
      if (!point) return;
      float ix[1];
      float iy[1];
      const float fx[1] = { point->xorigin() };
      const float fy[1] = { point->yorigin() };
      projectpts(pj, 1, fx, fy, ix, iy);
      point->insert_points(1, ix, iy);
    }
  break;
  case MapFeature::LineType:
    {
      MapEdge* edge = mf->edge_primitive();
      if (!edge) return;
      int npts = edge->npts();
      float ix[npts];
      float iy[npts];
      const float* fx = edge->xpoints();
      const float* fy = edge->ypoints();
      projectpts(pj, npts, fx, fy, ix, iy);
      edge->insert_points(npts, ix, iy);
    }
  break;
  case MapFeature::AreaType:
    {
      MapFace* face = mf->face_primitive();
      if (!face) return;
      int npts = face->npts();
      float ix[npts];
      float iy[npts];
      const float* fx = face->xpoints();
      const float* fy = face->ypoints();
      projectpts(pj, npts, fx, fy, ix, iy);
      face->insert_points(npts, ix, iy);
    }
  break;
  default:
    break;
  }
}

/*****************************************************************************/

ParamList* VpfFClassComp::_vpf_fcl_params = nil;

VpfFClassComp::VpfFClassComp(VpfFeatureClass* mfc) : MapFClassComp(mfc) {
    if (mfc->tiled()) {
	VpfFeatureClass* tileref = mfc->coverage()->library()->
	    coverage("tileref")->feature_class("tileref");
	_ntiles = tileref->nfeats();
	_tcomps = new OverlaysComp*[_ntiles+3];
	for (int i = 0; i < _ntiles+3; i++)
	    _tcomps[i] = nil;
    }
    else
	_ntiles = 0;
    _tmpname = nil;
}

VpfFClassComp::VpfFClassComp (istream& in) {
    _tmpname = nil;
    _valid = GetParamList()->read_args(in, this);
    Attribute* attr = GetAttributeList()->GetAttr("tile_name");
    const char* tilename = attr ? attr->Value()->string_ptr() : nil;
    if (tilename) {
      _mapfc = new VpfTiledFeatureClass(nil, _tmpname, tilename);
      _mapfc->tiled(true);
      _ntiles = 1;
      _tcomps = new OverlaysComp*[_ntiles+3];
      for (int i = 0; i < _ntiles+3; i++)
	_tcomps[i] = nil;
    } else {
      _mapfc = new VpfFeatureClass(nil, _tmpname);
    }
    delete _tmpname;
}

ParamList* VpfFClassComp::GetParamList() {
    if (!_vpf_fcl_params) 
	GrowParamList(_vpf_fcl_params = new ParamList());
    return _vpf_fcl_params;
}

void VpfFClassComp::GrowParamList(ParamList* pl) {
    pl->add_param("fclass", ParamStruct::required, &VpfFClassScript::ReadName, this, this);
    MapFClassComp::GrowParamList(pl);
}

VpfFClassComp::~VpfFClassComp() {
}

inline void fixpath(char* path) {
    int i = 0;
    while (path[i]) {
	if (path[i]=='\\')
	    path[i]='/';
	else {
	    int c = path[i];
	    c = tolower(path[i]);
	    path[i] = c;
	}
	i++;
    }
}

inline char *rjust( char *str )
{
   register int  len,i;

   len = strlen(str);
   i = len - 1;
   while ((i>0) && ((str[i]==0) || (str[i]==' '))) i--;
   if (i < (len-1)) str[i+1] = '\0';
   for (i=0;i<strlen(str);i++) if (str[i]=='\n') str[i] = '\0';
   return str;
}

void VpfFClassComp::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOWALL_CMD) || cmd->IsA(MAPSHOW_CMD)) {
	VpfEditor* editor = (VpfEditor*)cmd->GetEditor();
	if (editor)
	    editor->GetWindow()->cursor(hourglass);
	VpfExtent* ext = nil;
	int i;
	VpfCoverage* cov = (VpfCoverage*)_mapfc->coverage();
	VpfLibrary* lib = (VpfLibrary*)cov->library();
	VpfDatabase* db = (VpfDatabase*)lib->database();
	VpfCoverage* tcov = nil;
	VpfFeatureClass* tfc = nil;
	Attribute* tnattr = nil;
	if (_mapfc->tiled() && ( ((tcov = lib->coverage("tileref")) &&
	    (tfc = tcov->feature_class("tileref"))) ||
	    (tnattr = GetAttributeList()->GetAttr("tile_name")))
	) {
	    int ntiles;
	    Selection* sel = nil;
	    tnattr = GetAttributeList()->GetAttr("tile_name");
	    if (tnattr)
	      ntiles = 1;
	    else if (editor) {
	      sel = editor->tiles();
	      if (sel)
		ntiles = sel->Number();
	      else
		ntiles = 0;
	    }
	    else
	      ntiles = 1;
	    Iterator ti;
	    if (ntiles > 0) {
	        if (sel)
		  sel->First(ti);
		for (int j = 0; j < ntiles; j++) {
		    int tileid;
		    long int count;
		    VpfFeature* tile = nil;
		    const char* tname = nil;
		    if (sel) {
		      MapFeatureView* tview = (MapFeatureView*)sel->GetView(ti);
		      sel->Next(ti);
		      tile = (VpfFeature*)
			((MapFeatureComp*)tview->GetOverlayComp())->feature();
		    }
		    else if (tnattr) {
		      tname = tnattr->Value()->string_ptr();
		      ((VpfFeatureClass*)_mapfc)->tilename((char*)tname);
		    }
		    else {
		      int tid = ((MapShowCmd*)cmd)->GetTileId();
		      tile = tfc->feature(tid-1);
		    }
		    if (!tname)
		      tile->named_row_element("ID", &tileid, &count);
		    else
		      tileid = 0;
		    if (tname || !_tcomps[tileid]) {
		        char* tilename;
			if (tname)
			  tilename = (char*)tname;
			else {
			  tilename = (char*)tile
			    ->named_row_element("TILE_NAME", (void*)NULL, &count);
			  fixpath(tilename);
			  rjust(tilename);
			}
			if (!tnattr)
			  GetAttributeList()->add_attr
			    ("tile_name", new AttributeValue(tilename));

			_tcomps[tileid] = new OverlaysComp();
			if (ext == nil) {
			  if (tile)
			    ext = new VpfExtent(*tile->extent());
			}
			else
			    *ext = *ext + *tile->extent();

//			if (!tnattr) {
			VpfTiledFeatureClass* tfc = (VpfTiledFeatureClass*)_mapfc;
			if (strcmp(_mapfc->name(),"poarea")==0) {
			    for (i = 0; i < tfc->nfeats(tileid); i++) {
				MapFeature* vf = tfc->feature(tileid,i);
				if (vf->long_attr("POPYTYPE")==2) {
				    MapFeatureComp* mfc = new MapFeatureComp(vf);
				    ((VpfFeature*)vf)->graphic(mfc);
				    _tcomps[tileid]->Append(mfc);
				}
			    }
			    for (i = 0; i < tfc->nfeats(tileid); i++) {
				MapFeature* vf = tfc->feature(tileid,i);
				if (vf->long_attr("POPYTYPE")==1) {
				    MapFeatureComp* mfc = new MapFeatureComp(vf);
				    ((VpfFeature*)vf)->graphic(mfc);
				    _tcomps[tileid]->Append(mfc);
				}
			    }
			}
			else if (strcmp(_mapfc->name(),"hyarea")==0) {
			    for (int val = 1; val <= 6; val++) {
				for (i = 0; i < tfc->nfeats(tileid); i++) {
				    MapFeature* vf = tfc->feature(tileid,i);
				    if (vf->long_attr("HYPYTYPE")==val) {
					MapFeatureComp* mfc = new MapFeatureComp(vf);
					((VpfFeature*)vf)->graphic(mfc);
					_tcomps[tileid]->Append(mfc);
				    }
				}
			    }
			}
			else for (i = 0; i < tfc->nfeats(tileid); i++) {
			    MapFeature* vf = tfc->feature(tileid,i);
			    PJ* pj = ((MapShowCmd*)cmd)->GetProjection();
			    if (pj)
			      project_feature(pj, vf);
			    MapFeatureComp* mfc = new MapFeatureComp(vf);
			    ((VpfFeature*)vf)->graphic(mfc);

			    // create companion point comp for text comp
			    if (vf->type() == MapFeature::TextType) {
			      MapFeature* pf = new MapFeature
				(vf->feature_class(), -1, MapFeature::PointType);
			      MapPoint* pp = new MapPoint(pf);
			      float px = vf->text_primitive()->xorigin();
			      float py = vf->text_primitive()->yorigin();
			      pp->insert_points(1, &px, &py);
			      pf->point_primitive(pp);
			      MapFeatureComp* pfc = new MapFeatureComp(pf);
			      _tcomps[tileid]->Append(pfc);
			      mfc->pointcomp(pfc);
			    }

			    _tcomps[tileid]->Append(mfc);
			}
			Append(_tcomps[tileid]);
		    }
		}
	    }
	    else {
	      if (editor)
		editor->ExplainTileSel((char*)lib->name());
	    }
	}

	else if (strcmp(_mapfc->name(),"poarea")==0 && _comps->IsEmpty()) {
	    for (i = 0; i < _mapfc->nfeats(); i++) {
		MapFeature* vf = _mapfc->feature(i);
		if (vf->long_attr("POPYTYPE")==2) {
		    MapFeatureComp* mfc = new MapFeatureComp(vf);
		    ((VpfFeature*)vf)->graphic(mfc);
		    Append(mfc);
		}
	    }
	    for (i = 0; i < _mapfc->nfeats(); i++) {
		MapFeature* vf = _mapfc->feature(i);
		if (vf->long_attr("POPYTYPE")==1) {
		    MapFeatureComp* mfc = new MapFeatureComp(vf);
		    ((VpfFeature*)vf)->graphic(mfc);
		    Append(mfc);
		}
	    }
	}
	else if (strcmp(_mapfc->name(),"tileref")==0 && _comps->IsEmpty()) {
	    for (i = 1; i < _mapfc->nfeats(); i++) {
		MapFeature* vf = _mapfc->feature(i);
		MapFeatureComp* mfc = new MapFeatureComp(vf);
		((VpfFeature*)vf)->graphic(mfc);
		Append(mfc);
	    }
	}
	else if (_comps->IsEmpty()) {
	    // optional projected load (for nongraphic views)
	    PJ* pj = ((MapShowCmd*)cmd)->GetProjection();
	    for (i = 0; i < _mapfc->nfeats(); i++) {
		MapFeature* vf = _mapfc->feature(i);

		if (pj)
		    project_feature(pj, vf);

		MapFeatureComp* mfc = new MapFeatureComp(vf);
		((VpfFeature*)vf)->graphic(mfc);

		// create companion point comp for text comp
		if (vf->type() == MapFeature::TextType) {
		  MapFeature* pf = new MapFeature
		    (vf->feature_class(), -1, MapFeature::PointType);
		  MapPoint* pp = new MapPoint(pf);
		  float px = vf->text_primitive()->xorigin();
		  float py = vf->text_primitive()->yorigin();
		  pp->insert_points(1, &px, &py);
		  pf->point_primitive(pp);
		  MapFeatureComp* pfc = new MapFeatureComp(pf);
		  Append(pfc);
		  mfc->pointcomp(pfc);
		}

		Append(mfc);
	    }
	}
	if (editor && editor->Initializing()) {
	    float den, scale;
	    float x1, y1, x2, y2;
	    if (_mapfc->tiled()) {
	      double dx1, dy1, dx2, dy2;
	      int zone;
	      _mapfc->extent(dx1, dy1, dx2, dy2, zone);
	      x1 = dx1;
	      y1 = dy1;
	      x2 = dx2;
	      y2 = dy2;
	    }
	    else {
	      ext = ((VpfLibrary*)_mapfc->coverage()->library())->extent();
	      x1 = ext->x1;
	      y1 = ext->y1;
	      x2 = ext->x2;
	      y2 = ext->y2;
	    }
	    Catalog* catalog = unidraw->GetCatalog();
	    char* scl = strdup(catalog->GetAttribute("mapscale"));
	    den = atof(scl);
	    scale = 1.0/den;
	    VpfViewState* view = new VpfViewState(editor->GetProjection(),
						  x1, y1, scale);
	    editor->SetMapViewState(view);
	    PJ* proj = view->defproj();
	    UV data;
	    data.u = x1 * DEG_TO_RAD;
	    data.v = y1 * DEG_TO_RAD;
	    data = pj_fwd(data, proj);
	    x1 = data.u;
	    y1 = data.v;
	    data.u = x2 * DEG_TO_RAD;
	    data.v = y2 * DEG_TO_RAD;
	    data = pj_fwd(data, proj);
	    x2 = data.u;
	    y2 = data.v;
	    editor->GetViewer()->
		SetPage(new OverlayPage((x2-x1+1)/den,
					(y2-y1+1)/den));
	    editor->Initializing(false);
	}
	if (editor)
	    editor->FClassState((VpfFeatureClass*)_mapfc, true);
	Notify();
    }
    else if (cmd->IsA(MAPHIDE_CMD)) {
	VpfEditor* editor = (VpfEditor*)cmd->GetEditor();
	editor->GetWindow()->cursor(hourglass);
	MapFClassComp::Interpret(cmd);
	editor->FClassState((VpfFeatureClass*)_mapfc, false);
    }
    else
	MapFClassComp::Interpret(cmd);
}

void VpfFClassComp::Uninterpret(Command* cmd) {
}

Component* VpfFClassComp::Copy() { return new VpfFClassComp((VpfFeatureClass*)_mapfc); }
ClassId VpfFClassComp::GetClassId() { return VPFFCLASS_COMP; }
boolean VpfFClassComp::IsA(ClassId id) {
    return id == VPFFCLASS_COMP || MapFClassComp::IsA(id);
}

/*****************************************************************************/

VpfFClassView::VpfFClassView(VpfFClassComp* c) : MapFClassView(c) {
    _ignore_none = false;
    _hidden = false;
    _maxmag = MAXFLOAT;
}

void VpfFClassView::MaxMag(float f) {
    _maxmag = f;
}

float VpfFClassView::MaxMag() {
    return _maxmag;
}

void VpfFClassView::Hide() {
    _hidden = true;
    MapFClassView::Hide();
}

void VpfFClassView::Show() {
    if (strcmp(((VpfFClassComp*)GetOverlaysComp())->GetFeatureClass()->
	coverage()->library()->name(), "browse") == 0)
        _maxmag = 32;
    _hidden = false;
    float mag = GetViewer()->GetMagnification();
    if (mag <= _maxmag)
	MapFClassView::Show();
}

void VpfFClassView::AdjustForZoom (float factor, Coord cx, Coord cy) {
    if (strcmp(((VpfFClassComp*)GetOverlaysComp())->GetFeatureClass()->
	coverage()->library()->name(), "browse") == 0)
        _maxmag = 32;
    float mag = GetViewer()->GetMagnification();
    if (factor > 1.0 && mag >= _maxmag) {
        MapFClassView::Hide();
	Desensitize();
    }
    if (factor < 1.0 && mag <= (_maxmag * 2.0) && !_hidden) {
        MapFClassView::Show();
	Sensitize();
    }
    MapFClassView::AdjustForZoom(factor, cx, cy);
}

VpfFClassView::~VpfFClassView() {
}

void VpfFClassView::Update() {
  OverlaysView::Update();
}

ClassId VpfFClassView::GetClassId() { return VPFFCLASS_VIEW; }

boolean VpfFClassView::IsA(ClassId id) {
    return id == VPFFCLASS_VIEW || MapFClassView::IsA(id);
}

void VpfFClassView::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOWALL_CMD) || cmd->IsA(MAPSHOW_CMD)) {
	VpfEditor* editor = (VpfEditor*)cmd->GetEditor();
	VpfFeatureClass* mapfc = (VpfFeatureClass*)
	    ((VpfFClassComp*)GetOverlayComp())->GetFeatureClass();
	VpfCoverage* cov = (VpfCoverage*)mapfc->coverage();
	VpfLibrary* lib = (VpfLibrary*)cov->library();
	VpfDatabase* db = (VpfDatabase*)lib->database();
	VpfCoverage* tcov = lib->coverage("tileref");
	VpfFeatureClass* tfc = tcov->feature_class("tileref");
	Selection* sel = editor->tiles();
	if (sel) {
	int ntiles = sel->Number();
	Iterator ti;
	if (ntiles > 0) {
	    sel->First(ti);
	    for (int i = 0; i < ntiles; i++) {
		int tileid;
		long int count;
		MapFeatureView* tview = (MapFeatureView*)sel->GetView(ti);
		sel->Next(ti);
		VpfFeature* tile = (VpfFeature*)
		    ((MapFeatureComp*)tview->GetOverlayComp())->feature();
		tile->named_row_element("ID", &tileid, &count);
		VpfFClassComp* comp = (VpfFClassComp*)GetOverlaysComp();
		OverlaysComp** tcomps = comp->TileComps();
		Iterator vi;
		for (First(vi); !Done(vi); Next(vi)) {
		    OverlaysView* view = (OverlaysView*)GetView(vi);
		    if (view->GetOverlaysComp() == tcomps[tileid]
			&& view->Hidden()) {
			view->Show();
			view->Sensitize();
			GetViewer()->GetDamage()->Incur(view->GetGraphic());
			break;
		    }
		}
	    }
	    unidraw->Update();
	}
	}
    }
    else if (cmd->IsA(MAPHIDE_CMD)) {
	VpfEditor* editor = (VpfEditor*)cmd->GetEditor();
	VpfFeatureClass* mapfc = (VpfFeatureClass*)
	    ((VpfFClassComp*)GetOverlayComp())->GetFeatureClass();
	VpfCoverage* cov = (VpfCoverage*)mapfc->coverage();
	VpfLibrary* lib = (VpfLibrary*)cov->library();
	VpfDatabase* db = (VpfDatabase*)lib->database();
	VpfCoverage* tcov = lib->coverage("tileref");
	VpfFeatureClass* tfc = tcov->feature_class("tileref");
	Selection* sel = editor->tiles();
	if (sel) {
	int ntiles = sel->Number();
	Iterator ti;
	if (ntiles > 0) {
	    sel->First(ti);
	    for (int i = 0; i < ntiles; i++) {
		int tileid;
		long int count;
		MapFeatureView* tview = (MapFeatureView*)sel->GetView(ti);
		sel->Next(ti);
		VpfFeature* tile = (VpfFeature*)
		    ((MapFeatureComp*)tview->GetOverlayComp())->feature();
		tile->named_row_element("ID", &tileid, &count);
		VpfFClassComp* comp = (VpfFClassComp*)GetOverlaysComp();
		OverlaysComp** tcomps = comp->TileComps();
		Iterator vi;
		for (First(vi); !Done(vi); Next(vi)) {
		    OverlaysView* view = (OverlaysView*)GetView(vi);
		    if (view->GetOverlaysComp() == tcomps[tileid]
			&& !view->Hidden()) {
			GetViewer()->GetDamage()->Incur(view->GetGraphic());
			view->Hide();
			view->Desensitize();
			break;
		    }
		}
	    }
	    unidraw->Update();
	}
	}
    }
}

void VpfFClassView::Uninterpret(Command* cmd) {
}

Selection* VpfFClassView::SelectAll () {
    if (!((VpfFClassComp*)GetOverlayComp())->GetFeatureClass()->tiled())
	return MapFClassView::SelectAll();
    else {
	Iterator i;
	Selection* selection = MakeSelection();

	for (First(i); !Done(i); Next(i)) {
	    OverlayView* view = (OverlayView*) GetView(i);
	    if (!view->Desensitized()) 
		selection->Merge(view->SelectAll());
	}
	return selection;
    }
}

void VpfFClassView::IgnoreNone(boolean val) {
    _ignore_none = val;
}

Selection* VpfFClassView::ViewContaining (Coord x, Coord y) {
    if (!((VpfFClassComp*)GetOverlayComp())->GetFeatureClass()->tiled())
	return MapFClassView::ViewContaining(x, y);
    else {
	Iterator i;

	// speedup here by i-secting tiles?
	for (Last(i); !Done(i); Prev(i)) {
	    OverlayView* view = (OverlayView*) GetView(i);
	    if (!view->Desensitized()) {
		Selection* s = view->ViewContaining(x, y);
		if (!s->IsEmpty())
		    return s;
	    }
	}
	Selection* selection = MakeSelection();
	return selection;
    }
}

Selection* VpfFClassView::ViewsContaining (Coord x, Coord y) {
    if (!((VpfFClassComp*)GetOverlayComp())->GetFeatureClass()->tiled())
	return MapFClassView::ViewsContaining(x, y);
    else {
	Iterator i;
	Selection* s = MakeSelection();

	for (First(i); !Done(i); Next(i)) {
	    OverlayView* view = (OverlayView*) GetView(i);

	    if (!view->Desensitized()) {
		s->Merge(view->ViewsContaining(x, y));
	    }
	}
	return s;
    }
}

Selection* VpfFClassView::ViewIntersecting (
    Coord x0, Coord y0, Coord x1, Coord y1
) {
    if (!((VpfFClassComp*)GetOverlayComp())->GetFeatureClass()->tiled())
	return MapFClassView::ViewIntersecting(x0, y0, x1, y1);
    else {
	Iterator i;

	// speedup here by i-secting tiles?
	for (Last(i); !Done(i); Prev(i)) {
	    OverlayView* view = (OverlayView*) GetView(i);
	    if (!view->Desensitized()) {
		Selection* s = view->ViewIntersecting(x0, y0, x1, y1);
		if (!s->IsEmpty())
		    return s;
	    }
	}
	Selection* selection = MakeSelection();
	return selection;
    }
}

Selection* VpfFClassView::ViewsIntersecting (
    Coord x0, Coord y0, Coord x1, Coord y1
) {
    if (!((VpfFClassComp*)GetOverlayComp())->GetFeatureClass()->tiled()) {
	if (!_ignore_none)
	    return MapFClassView::ViewsIntersecting(x0, y0, x1, y1);
	else {
	    Iterator i;
	    Selection* s = MakeSelection();
	    BoxObj b(x0, y0, x1, y1);
	    PSPattern* save;
	    Catalog* catalog = unidraw->GetCatalog();
	    PSPattern* solid = catalog->FindGrayLevel(1.0);

	    for (First(i); !Done(i); Next(i)) {
		OverlayView* view = (OverlayView*) GetView(i);
		save = nil;
		if (view->GetGraphic()->GetPattern()->None()) {
		    save = view->GetGraphic()->GetPattern();
		    view->GetGraphic()->SetPattern(solid);
		}
		if (view->GetGraphic()->Intersects(b)) {
		    s->Append(view);
		}
		if (save)
		    view->GetGraphic()->SetPattern(save);
	    }
	    return s;
	}
    }
    else {
	Iterator i;
	Selection* s = MakeSelection();

	for (First(i); !Done(i); Next(i)) {
	    OverlayView* view = (OverlayView*) GetView(i);

	    if (!view->Desensitized()) {
		s->Merge(view->ViewsIntersecting(x0, y0, x1, y1));
	    }
	}
	return s;
    }
}

Selection* VpfFClassView::ViewsWithin (Coord x0, Coord y0, Coord x1, Coord y1){
    if (!((VpfFClassComp*)GetOverlayComp())->GetFeatureClass()->tiled())
	return MapFClassView::ViewsWithin(x0, y0, x1, y1);
    else {
	Iterator i;
	Selection* s = MakeSelection();

	for (First(i); !Done(i); Next(i)) {
	    OverlayView* view = (OverlayView*) GetView(i);

	    if (!view->Desensitized()) {
		s->Merge(view->ViewsWithin(x0, y0, x1, y1));
	    }
	}
	return s;
    }
}

/*****************************************************************************/

ParamList* VpfCovComp::_vpf_cov_params = nil;

VpfCovComp::VpfCovComp(VpfCoverage* vc) : MapCovComp((MapCoverage*)vc) {
}

VpfCovComp::VpfCovComp (istream& in) {
    _valid = GetParamList()->read_args(in, this);
   
}

ParamList* VpfCovComp::GetParamList() {
    if (!_vpf_cov_params) 
	GrowParamList(_vpf_cov_params = new ParamList());
    return _vpf_cov_params;
}

void VpfCovComp::GrowParamList(ParamList* pl) {
    pl->add_param("coverage", ParamStruct::required, &VpfCovScript::ReadName, this, &_mapcov);
    pl->add_param("fclasses", ParamStruct::required, &VpfCovScript::ReadChildren, this, this);
    MapCovComp::GrowParamList(pl);
}

VpfCovComp::~VpfCovComp() {
}

ClassId VpfCovComp::GetClassId() { return VPFCOV_COMP; }

boolean VpfCovComp::IsA(ClassId id) {
    return id == VPFCOV_COMP || MapCovComp::IsA(id);
}

MapFClassComp* VpfCovComp::CreateFClassComp(MapFeatureClass* mfc) {
    return new VpfFClassComp((VpfFeatureClass*)mfc);
}

void VpfCovComp::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOWALL_CMD)) {
	Iterator i;
	for (First(i); !Done(i); Next(i)) {
            GetComp(i)->Interpret(cmd);
        }
    } else if (cmd->IsA(MAPSHOW_CMD)) {
	MapFeatureClass* mapfc = (MapFeatureClass*)_mapcov->feature_class(((MapShowCmd*)cmd)->GetFClassName());
	if (mapfc && !mapfc->tiled())
	    MapCovComp::Interpret(cmd);
	else if (mapfc) {
	    // let both comp and view side interpret since may be different
	    // for each tile
	    MapFClassComp* mfc = FindFClassComp(mapfc);
	    if (!mfc) {
		mfc = CreateFClassComp(mapfc);
		Append(mfc);
		Notify();
	    }
	    mfc->Interpret(cmd);
            if (cmd->GetEditor())
	    ((MapEditor*)cmd->GetEditor())->GetMapDbView()->Interpret(cmd);
	}
    }
    else
	MapCovComp::Interpret(cmd);
}

/*****************************************************************************/

VpfCovView::VpfCovView(VpfCovComp* c) : MapCovView(c) {
}

ClassId VpfCovView::GetClassId() { return VPFCOV_VIEW; }

boolean VpfCovView::IsA(ClassId id) {
    return id == VPFCOV_VIEW || MapCovView::IsA(id);
}

void VpfCovView::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOWALL_CMD)) {
	Iterator i;
	for (First(i); !Done(i); Next(i)) {
            GetView(i)->Interpret(cmd);
        }
    } else if (cmd->IsA(MAPSHOW_CMD)) {
	MapFeatureClass* mapfc = (MapFeatureClass*)((MapCovComp*)GetGraphicComp())->GetCoverage()->feature_class(((MapShowCmd*)cmd)->GetFClassName());
	if (mapfc && !mapfc->tiled()) {
	    //MapCovView::Interpret(cmd);
	    VpfFClassView* mfv = (VpfFClassView*)FindFClassView(mapfc);
	    if (mfv /* && mfv->Hidden() */) {
		mfv->Show();
		mfv->Sensitize();
		GetViewer()->GetDamage()->Incur(GetGraphic());
		unidraw->Update();
	    }
	}
	else if (mapfc) {
	    VpfFClassView* mfv = (VpfFClassView*)FindFClassView(mapfc);
	    if (mfv)
		mfv->Interpret(cmd);
	}
    }
    else if (cmd->IsA(MAPHIDE_CMD)) {
	MapFeatureClass* mapfc = (MapFeatureClass*)((MapCovComp*)GetGraphicComp())->GetCoverage()->feature_class(((MapHideCmd*)cmd)->GetFClassName());
	if (mapfc && !mapfc->tiled()) {
	    //MapCovView::Interpret(cmd);
	    VpfFClassView* mfv = (VpfFClassView*)FindFClassView(mapfc);
	    if (mfv /* && !mfv->Hidden() */) {
		GetViewer()->GetDamage()->Incur(GetGraphic());
		mfv->Hide();
		mfv->Desensitize();
		unidraw->Update();
	    }
	}
	else if (mapfc) {
	    VpfFClassView* mfv = (VpfFClassView*)FindFClassView(mapfc);
	    if (mfv)
		mfv->Interpret(cmd);
	}
    }
}

/*****************************************************************************/

ParamList* VpfLibComp::_vpf_lib_params = nil;

VpfLibComp::VpfLibComp(VpfLibrary* vl) : MapLibComp(vl) {
}

VpfLibComp::VpfLibComp (istream& in) {
    _valid = GetParamList()->read_args(in, this);
}

ParamList* VpfLibComp::GetParamList() {
    if (!_vpf_lib_params) 
	GrowParamList(_vpf_lib_params = new ParamList());
    return _vpf_lib_params;
}

void VpfLibComp::GrowParamList(ParamList* pl) {
    pl->add_param("library", ParamStruct::required, &VpfLibScript::ReadName, this, &_maplib);
    pl->add_param("coverages", ParamStruct::required, &VpfLibScript::ReadChildren, this, this);
    MapLibComp::GrowParamList(pl);
}

VpfLibComp::~VpfLibComp() {
}

MapCovComp* VpfLibComp::CreateCovComp(MapCoverage* mc) {
    return new VpfCovComp((VpfCoverage*)mc);
}

/*****************************************************************************/

ParamList* VpfDbComp::_vpf_db_params = nil;

VpfDbComp::VpfDbComp(VpfDatabase* vdb) : MapDbComp((MapDatabase*)vdb) {
}

VpfDbComp::VpfDbComp (istream& in) {
    _valid = GetParamList()->read_args(in, this);
}

void VpfDbComp::init() {
    VpfDatabase* db = GetDatabase();
    Iterator i, j, k;

    //initialize child libraries, coverages and feature classes
    for (First(i); !Done(i); Next(i)) {
        VpfLibComp* libcomp = (VpfLibComp*)GetComp(i);
	VpfLibrary* lib = libcomp->GetLibrary();
        lib->database(db);

        for (libcomp->First(j); !libcomp->Done(j); libcomp->Next(j)) {
            VpfCovComp* covcomp = (VpfCovComp*)libcomp->GetComp(j);
	    VpfCoverage* cov = covcomp->GetCoverage();
            cov->library(lib);

	    for (covcomp->First(k); !covcomp->Done(k); covcomp->Next(k)) {
                VpfFClassComp* fclcomp = (VpfFClassComp*)covcomp->GetComp(k);
	        VpfFeatureClass* fcl = fclcomp->GetFeatureClass();
                fcl->coverage(cov);
            }
        }
    }
}

void VpfDbComp::Configure(Editor* ed) {
  init();
}

ParamList* VpfDbComp::GetParamList() {
    if (!_vpf_db_params) 
	GrowParamList(_vpf_db_params = new ParamList());
    return _vpf_db_params;
}

void VpfDbComp::GrowParamList(ParamList* pl) {
    pl->add_param("path", ParamStruct::required, &VpfDbScript::ReadPath, this, &_mapdb);
    pl->add_param("libraries", ParamStruct::required, &VpfDbScript::ReadChildren, this, this);
    MapDbComp::GrowParamList(pl);
}

VpfDbComp::~VpfDbComp() {
}

MapLibComp* VpfDbComp::CreateLibComp(MapLibrary* ml) {
    return new VpfLibComp((VpfLibrary*)ml);
}

/*****************************************************************************/

ParamList* VpfIdrawComp::_vpf_idraw_params = nil;

VpfIdrawComp::VpfIdrawComp () : MapIdrawComp() { }
ClassId VpfIdrawComp::GetClassId () { return VPF_IDRAW_COMP; }

boolean VpfIdrawComp::IsA (ClassId id) {
    return VPF_IDRAW_COMP == id || MapIdrawComp::IsA(id);
}

VpfIdrawComp::VpfIdrawComp (istream& in) {
    _valid = GetParamList()->read_args(in, this);
}

ParamList* VpfIdrawComp::GetParamList() {
    if (!_vpf_idraw_params) 
	GrowParamList(_vpf_idraw_params = new ParamList());
    return _vpf_idraw_params;
}

void VpfIdrawComp::GrowParamList(ParamList* pl) {
    pl->add_param("databases", ParamStruct::required, &VpfIdrawScript::ReadChildren, this, this);
    pl->add_param("projection", ParamStruct::keyword, &VpfIdrawScript::ReadProjection, this, this);
    MapIdrawComp::GrowParamList(pl); 
}

Component* VpfIdrawComp::Copy () {
    return new VpfIdrawComp();
}
